from dataclasses import dataclass
from typing import List, Optional, Union

from .client_portfolio import CashPortfolio
from .constraints import ConstraintFactory
from .full_optimizer_node import FullSpecOptimizationNode, GenericObjectiveFunction, TaxOptimizationSetting, \
    OptimizationSettings, TaxArbitrage, CashFlowOptSetting, DoNotTradeList, DoNotTradeExpression
from .mos_config import Strategy, SimulationSettings, ReferenceUniverse
from .profile import Profile
from .utils.enums import CalculationTypeEnum


@dataclass
class OptimizationOptions:
    """
    Dataclass for providing options for creating generic optimization node.

    Args:
        solver (int): to be tested, should be the same as values used in
            opstool. {
                BARRAOPT::eNO_SOLVER = 0,
                BARRAOPT::eQUADRATIC_SOLVER = 1,
                BARRAOPT::eSECOND_ORDER_CONE_SOLVER = 2,
                BARRAOPT::eNONLINEAR_SOLVER = 3
            }
        paring_approach (int): to be tested. Should be the same as options that
            can be entered in open optimizer.
        transaction_type (str): valid values are [allowAll, buyNone, sellNone,
            shortNone, buyFromUniv, sellNoneBuyFromUniv, buyShortFromUniv,
            disallowBuyShort, disallowSellCover,
            buyShortUnivNoSellCover, buyNoneSellFromUniv, univOnly]
        alpha_attribute (str): attribute field name that will be used as alpha
            values for investable
        return_multiplier (float,int): multiplier for return term.

    """

    solver: Optional[int] = None
    paring_approach: Optional[int] = None
    transaction_type: str = "allowAll"
    alpha_attribute: Optional[str] = None
    return_multiplier: Optional[Union[float,int]] = None


class GenericTaxTemplate(Profile):
    """
    Generic template to use for all tax optimization strategies.

    Args:
        calculation_type (CalculationTypeEnum): (optional) Use to specify behavior of the calculation. The following options are available:
        
                    •REBALANCE - (default)rebalance a portfolio for one day only. Use the latest available portfolio given the date of calculation.
                    •SIMULATION - compute a time series of portfolio using the latest available portfolio given the first day of calculation. Skip last rebalancing.
                    •EOD/BACKCALCULATION - compute a time series of portfolio using the portfolio at the day of the calculation. Fail if not present. Perform last rebalancing.
                    
        te_limit (float): Tracking error limit (predicted active risk) set in decimal terms. Default value is None.
        te_limit_soft (bool): If True, Tracking Error limit is a soft constraint. If False, Tracking Error limit is a hard constraint. Default value is True.
        max_assets (int): Maximum number of assets in optimal portfolio, soft bound. Default value is None.
        analysis_date (str): The date in YYYY-MM-DD format for which the account is being REBALANCED. Do not set with from_date.
        from_date (str): The starting date in YYYY-MM-DD format for a SIMULATION/ BACKCALCULATION. Do not set with analysis_date.
        to_date (str): The ending date in YYYY-MM-DD format for a SIMULATION/ BACKCALCULATION. Do not set with analysis_date.
        benchmark_list (list): List of MSCI index to use as a benchmark. Default value is None.
        universe_list (list): List of MSCI index to use as a universe. Default value is None.
        portfolio (Portfolio):  A portfolio object to use as the initial portfolio of a rebalance or simulation. Default value is CashPortfolio with 10,000,000 USD.
        to_limit (int) : Maximum turnover the optimizer must observe in producing an optimal portfolio.
        risk_aversion (float): Risk aversion to use in the optimization problem. Default value is 0.0075.
        tax_multiplier (float): Tax penalty rate from gains in the objective function. Default value is 1.
        loss_benefit_multiplier (float): Tax benefit rate from losses in the objective function. Default value is 1.
        prefix_nodes (list): Nodes to screen the  ahead of the optimization, such as BISR screens of Node. Default value is None.
        tax_arbitrage (List[TaxArbitrage]): Tax arbitrage ranges. Default value is None.
        tax_unit (str): Tax unit, "dollar" or "decimal". Default value is "dollar".
        no_trade_criteria (DoNotTradeList, DoNotTradeExpression): Do Not Trade criteria from a previously uploaded portfolio. Default value is None.
        no_hold_criteria (DoNotTradeList, DoNotTradeExpression): Do Not Hold criteria from a previously uploaded portfolio. Default value is None.
        no_sell_criteria (DoNotTradeList, DoNotTradeExpression): No Sell criteria from a previously uploaded portfolio. Default value is None.
        no_buy_criteria (DoNotTradeList, DoNotTradeExpression): No Buy criteria from a previously uploaded portfolio. Default value is None.
        profile_name (str): Name of the profile request. Useful in case of bulk jobs.

    Returns:
            None

    """

    def __init__(
            self,
            calculation_type=CalculationTypeEnum.REBALANCE,
            te_limit=None,
            te_limit_soft=True,
            max_assets=None,
            analysis_date=None,
            from_date=None,
            to_date=None,
            benchmark_list: List = None,
            universe_list: List = None,
            portfolio=CashPortfolio(),
            to_limit=None,
            risk_aversion=0.01,
            tax_multiplier=1,
            loss_benefit_multiplier=0,
            prefix_nodes=None,
            tax_arbitrage: List[TaxArbitrage] = None,
            tax_unit: str = "dollar",
            cash_upper_bound: float = 0.010,
            cash_lower_bound: float = 0.009,
            constraints: List = None,
            opt_options: OptimizationOptions = OptimizationOptions(),
            no_trade_criteria: Union[DoNotTradeExpression, DoNotTradeList] = None,
            no_hold_criteria: Union[DoNotTradeExpression, DoNotTradeList] = None,
            no_sell_criteria: Union[DoNotTradeExpression, DoNotTradeList] = None,
            no_buy_criteria: Union[DoNotTradeExpression, DoNotTradeList] = None,
            profile_name: str = None
    ):
        if analysis_date is None:
            if from_date is not None and to_date is not None:
                calculation_type = CalculationTypeEnum.SIMULATION

        # if isinstance(portfolio, CashPortfolio):
        cashflow_usd = 0  # cash already inline
        # else:
        #     cashflow_usd = portfolio.initial_cash # for now, apply it to the optimization, in the future this should be added to omps

        if constraints is None:
            constraints = []

        if te_limit:
            constraints += [ConstraintFactory.RiskConstraint(upper_bound=te_limit,
                                                             is_soft=te_limit_soft,
                                                             reference_portfolio="BaseBenchmark")]
        if max_assets:
            constraints += [ConstraintFactory.NumberOfAssets(max=max_assets)]
        if to_limit:
            constraints += [ConstraintFactory.TurnoverConstraint(upper_bound=to_limit)]

        objective_function = GenericObjectiveFunction(
            tax_term=tax_multiplier,
            loss_benefit_term=loss_benefit_multiplier,
            cf_risk_aversion=risk_aversion,
            sp_risk_aversion=risk_aversion,
            alpha_attribute=opt_options.alpha_attribute,
            alpha_term=opt_options.return_multiplier
        )
        tax_opt_setting = TaxOptimizationSetting(tax_unit=tax_unit, selling_order_rule="hifo",
                                                 long_term_tax_rate=0.238, short_term_tax_rate=0.408,
                                                 wash_sale_rule="tradeoff")
        # if cashflow_usd:
        #     cash_flow_settings = CashFlowOptSetting(amount=cashflow_usd,cash_type='AMOUNT')
        # else:
        cash_flow_settings = None
        opt_settings = OptimizationSettings(
            tax_optimization_setting=tax_opt_setting,
            cash_flow_settings=cash_flow_settings,
            cash_in_portfolio_upper_bound=cash_upper_bound,
            cash_in_portfolio_lower_bound=cash_lower_bound,
            paring_approach=opt_options.paring_approach,
            solver=opt_options.solver,
            transaction_type=opt_options.transaction_type
        )

        tax_opt_node = FullSpecOptimizationNode(
            constraints=constraints,
            objective_function=objective_function,
            tax_arbitrages=tax_arbitrage,
            opt_settings=opt_settings,
            do_not_trade_criteria=no_trade_criteria,
            do_not_hold_criteria=no_hold_criteria,
            no_sell=no_sell_criteria,
            no_buy=no_buy_criteria,
        )
        ref_universe = ReferenceUniverse(
            benchmark=benchmark_list,
            universe=universe_list,
            portfolio=portfolio,
        )
        if prefix_nodes:
            node_list = prefix_nodes + [tax_opt_node]
        else:
            node_list = [tax_opt_node]
        strat = Strategy(node_list=node_list, ref_universe=ref_universe)
        sim_settings = SimulationSettings(calculation_type=calculation_type,
                                          from_date=from_date, to_date=to_date,
                                          analysis_date=analysis_date)

        super().__init__(
            strategy=strat,
            simulation_settings=sim_settings,
            profile_name=profile_name
        )


class TaxAdvantagedModelTrackingTemplate(GenericTaxTemplate):
    """
    Template to maximize losses while managing risk relative to a benchmark with risk aversion and tracking error limit.

    Args:
        calculation_type (CalculationTypeEnum): (optional) Use to specify behavior of the calculation. The following options are available:
            
                    •REBALANCE - (default)rebalance a portfolio for one day only. Use the latest available portfolio given the date of calculation.
                    •SIMULATION - compute a time series of portfolio using the latest available portfolio given the first day of calculation. Skip last rebalancing.
                    •EOD/BACKCALCULATION - compute a time series of portfolio using the portfolio at the day of the calculation. Fail if not present. Perform last rebalancing.
                    
        te_limit (float): Tracking error limit (predicted active risk) set in decimal terms. Default value is 0.005.
        te_limit_soft (bool): If True, Tracking Error limit is a soft constraint. If False, Tracking Error limit is a hard constraint. Default value is True.
        max_assets (int): Maximum number of assets in optimal portfolio, soft bound. Default value is None.
        analysis_date (str): The date in YYYY-MM-DD format for which the account is being REBALANCED. Do not set with from_date.
        from_date (str): The starting date in YYYY-MM-DD format for a SIMULATION/ BACKCALCULATION. Do not set with analysis_date.
        to_date (str): The ending date in YYYY-MM-DD format for a SIMULATION/ BACKCALCULATION. Do not set with analysis_date.
        benchmark_list (list): List of MSCI index to use as a benchmark. Default value is None.
        universe_list (list): List of MSCI index to use as a universe. Default value is None.
        portfolio (Portfolio):  A portfolio object to use as the initial portfolio of a rebalance or simulation. Default value is CashPortfolio with 10,000,000 USD.
        to_limit (int) : Maximum turnover the optimizer must observe in producing an optimal portfolio.
        risk_aversion (float): Risk aversion to use in the optimization problem. Default value is 0.0001.
        tax_multiplier (float): Tax penalty rate from gains in the objective function. Default value is 1.
        loss_benefit_multiplier (float): Tax benefit rate from losses in the objective function. Default value is 1.
        prefix_nodes (list): Nodes to screen the  ahead of the optimization, such as BISR screens of Node. Default value is None.
        tax_arbitrage (List[TaxArbitrage]): Tax arbitrage ranges. Default value is None.
        tax_unit (str): Tax unit, "dollar" or "decimal". Default value is "dollar".
        no_trade_criteria (DoNotTradeList, DoNotTradeExpression): Do Not Trade criteria from a previously uploaded portfolio. Default value is None.
        no_hold_criteria (DoNotTradeList, DoNotTradeExpression): Do Not Hold criteria from a previously uploaded portfolio. Default value is None.
        no_sell_criteria (DoNotTradeList, DoNotTradeExpression): No Sell criteria from a previously uploaded portfolio. Default value is None.
        no_buy_criteria (DoNotTradeList, DoNotTradeExpression): No Buy criteria from a previously uploaded portfolio. Default value is None.
        profile_name (str): Name of the profile request. Useful in case of bulk jobs.
    """

    def __init__(self, risk_aversion=0.0001, tax_multiplier=1, loss_benefit_multiplier=1, te_limit=0.005,
                 te_limit_soft=True, **kwargs):
        super().__init__(risk_aversion=risk_aversion, tax_multiplier=tax_multiplier,
                         loss_benefit_multiplier=loss_benefit_multiplier, te_limit=te_limit,
                         te_limit_soft=te_limit_soft, **kwargs)


class TaxNeutralTemplate(GenericTaxTemplate):
    """
    Template to stay tax neutral (gains similar to losses) while managing risk relative to a benchmark with risk aversion and tracking error limit.

    Args:
        calculation_type (CalculationTypeEnum): (optional) Use to specify behavior of the calculation. The following options are available:
            
                    •REBALANCE - (default)rebalance a portfolio for one day only. Use the latest available portfolio given the date of calculation.
                    •SIMULATION - compute a time series of portfolio using the latest available portfolio given the first day of calculation. Skip last rebalancing.
                    •EOD/BACKCALCULATION - compute a time series of portfolio using the portfolio at the day of the calculation. Fail if not present. Perform last rebalancing.
                    
        te_limit (float): Tracking error limit (predicted active risk) set in decimal terms. Default value is None.
        te_limit_soft (bool): If True, Tracking Error limit is a soft constraint. If False, Tracking Error limit is a hard constraint. Default value is True.
        max_assets (int): Maximum number of assets in optimal portfolio, soft bound. Default value is None.
        analysis_date (str): The date in YYYY-MM-DD format for which the account is being REBALANCED. Do not set with from_date.
        from_date (str): The starting date in YYYY-MM-DD format for a SIMULATION/ BACKCALCULATION. Do not set with analysis_date.
        to_date (str): The ending date in YYYY-MM-DD format for a SIMULATION/ BACKCALCULATION. Do not set with analysis_date.
        benchmark_list (list): List of MSCI index to use as a benchmark. Default value is None.
        universe_list (list): List of MSCI index to use as a universe. Default value is None.
        portfolio (Portfolio):  A portfolio object to use as the initial portfolio of a rebalance or simulation. Default value is CashPortfolio with 10,000,000 USD.
        to_limit (int) : Maximum turnover the optimizer must observe in producing an optimal portfolio.
        risk_aversion (float): Risk aversion to use in the optimization problem. Default value is 1.
        tax_multiplier (float): Tax penalty rate from gains in the objective function. Default value is 100.
        loss_benefit_multiplier (float): Tax benefit rate from losses in the objective function. Default value is 0.0001.
        prefix_nodes (list): Nodes to screen the  ahead of the optimization, such as BISR screens of Node. Default value is None.
        tax_arbitrage (List[TaxArbitrage]): Tax arbitrage ranges. Default value is None.
        tax_unit (str): Tax unit, "dollar" or "decimal". Default value is "dollar".
        no_trade_criteria (DoNotTradeList, DoNotTradeExpression): Do Not Trade criteria from a previously uploaded portfolio. Default value is None.
        no_hold_criteria (DoNotTradeList, DoNotTradeExpression): Do Not Hold criteria from a previously uploaded portfolio. Default value is None.
        no_sell_criteria (DoNotTradeList, DoNotTradeExpression): No Sell criteria from a previously uploaded portfolio. Default value is None.
        no_buy_criteria (DoNotTradeList, DoNotTradeExpression): No Buy criteria from a previously uploaded portfolio. Default value is None.
        profile_name (str): Name of the profile request. Useful in case of bulk jobs.
    """

    def __init__(self, risk_aversion=1, tax_multiplier=100, loss_benefit_multiplier=0.0001, use_tax_constraint=False,
                 **kwargs):
        tax_arbitrage = [TaxArbitrage(upper_bound=0, lower_bound=None)] if use_tax_constraint else []
        super().__init__(risk_aversion=risk_aversion, tax_multiplier=tax_multiplier,
                         loss_benefit_multiplier=loss_benefit_multiplier,
                         tax_arbitrage=tax_arbitrage, **kwargs)


class GainBudgetingTemplate(GenericTaxTemplate):
    """
    Template to stay as close as possible to a benchmark while staying within a user-specified gain budget.

    Args:
        calculation_type (CalculationTypeEnum): (optional) Use to specify behavior of the calculation. The following options are available:
            
                    •REBALANCE - (default)rebalance a portfolio for one day only. Use the latest available portfolio given the date of calculation.
                    •SIMULATION - compute a time series of portfolio using the latest available portfolio given the first day of calculation. Skip last rebalancing.
                    •EOD/BACKCALCULATION - compute a time series of portfolio using the portfolio at the day of the calculation. Fail if not present. Perform last rebalancing.
                    
        te_limit (float): Tracking error limit (predicted active risk) set in decimal terms. Default value is None.
        te_limit_soft (bool): If True, Tracking Error limit is a soft constraint. If False, Tracking Error limit is a hard constraint. Default value is True.
        max_assets (int): Maximum number of assets in optimal portfolio, soft bound. Default value is None.
        analysis_date (str): The date in YYYY-MM-DD format for which the account is being REBALANCED. Do not set with from_date.
        from_date (str): The starting date in YYYY-MM-DD format for a SIMULATION/ BACKCALCULATION. Do not set with analysis_date.
        to_date (str): The ending date in YYYY-MM-DD format for a SIMULATION/ BACKCALCULATION. Do not set with analysis_date.
        benchmark_list (list): List of MSCI index to use as a benchmark. Default value is None.
        universe_list (list): List of MSCI index to use as a universe. Default value is None.
        portfolio (Portfolio):  A portfolio object to use as the initial portfolio of a rebalance or simulation. Default value is CashPortfolio with 10,000,000 USD.
        to_limit (int) : Maximum turnover the optimizer must observe in producing an optimal portfolio.
        risk_aversion (float): Risk aversion to use in the optimization problem. Default value is 100.
        tax_multiplier (float): Tax penalty rate from gains in the objective function. Default value is 1.
        loss_benefit_multiplier (float): Tax benefit rate from losses in the objective function. Default value is 1.
        prefix_nodes (list): Nodes to screen the  ahead of the optimization, such as BISR screens of Node. Default value is None.
        tax_arbitrage (List[TaxArbitrage]): Tax arbitrage ranges. Default value is None.
        tax_unit (str): Tax unit, "dollar" or "decimal". Default value is "dollar".
        gain_budget (float): maximum allowed gains in dollar value. Default value is None.
        no_trade_criteria (DoNotTradeList, DoNotTradeExpression): Do Not Trade criteria from a previously uploaded portfolio. Default value is None.
        no_hold_criteria (DoNotTradeList, DoNotTradeExpression): Do Not Hold criteria from a previously uploaded portfolio. Default value is None.
        no_sell_criteria (DoNotTradeList, DoNotTradeExpression): No Sell criteria from a previously uploaded portfolio. Default value is None.
        no_buy_criteria (DoNotTradeList, DoNotTradeExpression): No Buy criteria from a previously uploaded portfolio. Default value is None.
        profile_name (str): Name of the profile request. Useful in case of bulk jobs.
    """

    def __init__(self, gain_budget: float, risk_aversion: float = 100, tax_multiplier: float = 1,
                 loss_benefit_multiplier: float = 1, tax_unit="dollar", **kwargs):
        tax_arbitrage = [TaxArbitrage(upper_bound=gain_budget)]
        super().__init__(risk_aversion=risk_aversion, tax_multiplier=tax_multiplier,
                         loss_benefit_multiplier=loss_benefit_multiplier,
                         tax_unit=tax_unit, tax_arbitrage=tax_arbitrage, **kwargs)


class MaxLossHarvestingTemplate(GenericTaxTemplate):
    """
    Template to maximize losses without aversion to risk beyond a user specified optional bound.

    Args:
        calculation_type (CalculationTypeEnum): (optional) Use to specify behavior of the calculation. The following options are available:
            
                    •REBALANCE - (default)rebalance a portfolio for one day only. Use the latest available portfolio given the date of calculation.
                    •SIMULATION - compute a time series of portfolio using the latest available portfolio given the first day of calculation. Skip last rebalancing.
                    •EOD/BACKCALCULATION - compute a time series of portfolio using the portfolio at the day of the calculation. Fail if not present. Perform last rebalancing.
                    
        te_limit (float): Tracking error limit (predicted active risk) set in decimal terms. Default value is None.
        te_limit_soft (bool): If True, Tracking Error limit is a soft constraint. If False, Tracking Error limit is a hard constraint. Default value is True.
        max_assets (int): Maximum number of assets in optimal portfolio, soft bound. Default value is None.
        analysis_date (str): The date in YYYY-MM-DD format for which the account is being REBALANCED. Do not set with from_date.
        from_date (str): The starting date in YYYY-MM-DD format for a SIMULATION/ BACKCALCULATION. Do not set with analysis_date.
        to_date (str): The ending date in YYYY-MM-DD format for a SIMULATION/ BACKCALCULATION. Do not set with analysis_date.
        benchmark_list (list): List of MSCI index to use as a benchmark. Default value is None.
        universe_list (list): List of MSCI index to use as a universe. Default value is None.
        portfolio (Portfolio):  A portfolio object to use as the initial portfolio of a rebalance or simulation. Default value is CashPortfolio with 10,000,000 USD.
        to_limit (int) : Maximum turnover the optimizer must observe in producing an optimal portfolio.
        risk_aversion (float): Risk aversion to use in the optimization problem. Default value is 0.0001.
        tax_multiplier (float): Tax penalty rate from gains in the objective function. Default value is 1.
        loss_benefit_multiplier (float): Tax benefit rate from losses in the objective function. Default value is 1.
        prefix_nodes (list): Nodes to screen the  ahead of the optimization, such as BISR screens of Node. Default value is None.
        tax_arbitrage (List[TaxArbitrage]): Tax arbitrage ranges. Default value is None.
        tax_unit (str): Tax unit, "dollar" or "decimal". Default value is "dollar".
        no_trade_criteria (DoNotTradeList, DoNotTradeExpression): Do Not Trade criteria from a previously uploaded portfolio. Default value is None.
        no_hold_criteria (DoNotTradeList, DoNotTradeExpression): Do Not Hold criteria from a previously uploaded portfolio. Default value is None.
        no_sell_criteria (DoNotTradeList, DoNotTradeExpression): No Sell criteria from a previously uploaded portfolio. Default value is None.
        no_buy_criteria (DoNotTradeList, DoNotTradeExpression): No Buy criteria from a previously uploaded portfolio. Default value is None.
        profile_name (str): Name of the profile request. Useful in case of bulk jobs.
    """

    def __init__(self, risk_aversion: float = 0.0001, tax_multiplier=1, loss_benefit_multiplier=1, **kwargs):
        super().__init__(risk_aversion=risk_aversion, tax_multiplier=tax_multiplier,
                         loss_benefit_multiplier=loss_benefit_multiplier, **kwargs)


class GainRealizationTemplate(GenericTaxTemplate):
    """
    Template to maximize gains while managing risk relative to a benchmark with risk aversion and tracking error limit.

    Args:
        calculation_type (CalculationTypeEnum): (optional) Use to specify behavior of the calculation. The following options are available:
            
                    •REBALANCE - (default)rebalance a portfolio for one day only. Use the latest available portfolio given the date of calculation.
                    •SIMULATION - compute a time series of portfolio using the latest available portfolio given the first day of calculation. Skip last rebalancing.
                    •EOD/BACKCALCULATION - compute a time series of portfolio using the portfolio at the day of the calculation. Fail if not present. Perform last rebalancing.
                    
        te_limit (float): Tracking error limit (predicted active risk) set in decimal terms. Default value is None.
        te_limit_soft (bool): If True, Tracking Error limit is a soft constraint. If False, Tracking Error limit is a hard constraint. Default value is True.
        max_assets (int): Maximum number of assets in optimal portfolio, soft bound. Default value is None.
        analysis_date (str): The date in YYYY-MM-DD format for which the account is being REBALANCED. Do not set with from_date.
        from_date (str): The starting date in YYYY-MM-DD format for a SIMULATION/ BACKCALCULATION. Do not set with analysis_date.
        to_date (str): The ending date in YYYY-MM-DD format for a SIMULATION/ BACKCALCULATION. Do not set with analysis_date.
        benchmark_list (list): List of MSCI index to use as a benchmark. Default value is None.
        universe_list (list): List of MSCI index to use as a universe. Default value is None.
        portfolio (Portfolio):  A portfolio object to use as the initial portfolio of a rebalance or simulation. Default value is CashPortfolio with 10,000,000 USD.
        to_limit (int) : Maximum turnover the optimizer must observe in producing an optimal portfolio.
        risk_aversion (float): Risk aversion to use in the optimization problem. Default value is 0.0001.
        tax_multiplier (float): Tax penalty rate from gains in the objective function. Default value is -1.
        loss_benefit_multiplier (float): Tax benefit rate from losses in the objective function. Default value is -1.
        prefix_nodes (list): Nodes to screen the  ahead of the optimization, such as BISR screens of Node. Default value is None.
        tax_arbitrage (List[TaxArbitrage]): Tax arbitrage ranges. Default value is None.
        tax_unit (str): Tax unit, "dollar" or "decimal". Default value is "dollar".
        no_trade_criteria (DoNotTradeList, DoNotTradeExpression): Do Not Trade criteria from a previously uploaded portfolio. Default value is None.
        no_hold_criteria (DoNotTradeList, DoNotTradeExpression): Do Not Hold criteria from a previously uploaded portfolio. Default value is None.
        no_sell_criteria (DoNotTradeList, DoNotTradeExpression): No Sell criteria from a previously uploaded portfolio. Default value is None.
        no_buy_criteria (DoNotTradeList, DoNotTradeExpression): No Buy criteria from a previously uploaded portfolio. Default value is None.
        profile_name (str): Name of the profile request. Useful in case of bulk jobs.
    """

    def __init__(self, risk_aversion: float = 0.0001, tax_multiplier=-1, loss_benefit_multiplier=-1, **kwargs):
        super().__init__(risk_aversion=risk_aversion, tax_multiplier=tax_multiplier,
                         loss_benefit_multiplier=loss_benefit_multiplier, **kwargs)


class PureOptimizationTemplate(Profile):
    """
    Generic template to use for all tax optimization strategies
    """

    def __init__(
            self,
            calculation_type=CalculationTypeEnum.REBALANCE,
            te_limit=None,
            te_limit_soft=True,
            max_assets=None,
            analysis_date=None,
            from_date=None,
            to_date=None,
            benchmark_list: List = None,
            universe_list: List = None,
            portfolio=CashPortfolio(),
            to_limit=None,
            risk_aversion=0.01,
            prefix_nodes=None,
            cash_upper_bound: float = 0.010,
            cash_lower_bound: float = 0.009,
            constraints: List = None,
            opt_options: OptimizationOptions = OptimizationOptions(),
            profile_name: str = None
    ):

        # if isinstance(portfolio, CashPortfolio):
        cashflow_usd = 0  # cash already inline
        # else:
        #     cashflow_usd = portfolio.initial_cash  # for now, apply it to the optimization, in the future this should be added to omps

        if constraints is None:
            constraints = []
        if te_limit:
            constraints += [ConstraintFactory.RiskConstraint(upper_bound=te_limit,
                                                             is_soft=te_limit_soft,
                                                             reference_portfolio="BaseBenchmark")]
        if max_assets:
            constraints += [ConstraintFactory.NumberOfAssets(max=max_assets)]
        if to_limit:
            constraints += [ConstraintFactory.TurnoverConstraint(to_limit)]

        objective_function = GenericObjectiveFunction(
            cf_risk_aversion=risk_aversion,
            sp_risk_aversion=risk_aversion,
            alpha_attribute=opt_options.alpha_attribute,
            alpha_term=opt_options.return_multiplier
        )
        tax_opt_setting = None
        cash_flow_settings = CashFlowOptSetting(amount=cashflow_usd, cash_type='AMOUNT')
        opt_settings = OptimizationSettings(
            tax_optimization_setting=tax_opt_setting,
            cash_flow_settings=cash_flow_settings,
            cash_in_portfolio_upper_bound=cash_upper_bound,
            cash_in_portfolio_lower_bound=cash_lower_bound,
            paring_approach=opt_options.paring_approach,
            solver=opt_options.solver,
            transaction_type=opt_options.transaction_type
        )
        tax_opt_node = FullSpecOptimizationNode(
            constraints=constraints,
            objective_function=objective_function,
            opt_settings=opt_settings
        )
        ref_universe = ReferenceUniverse(
            benchmark=benchmark_list,
            universe=universe_list,
            portfolio=portfolio,
        )
        if prefix_nodes:
            node_list = prefix_nodes + [tax_opt_node]
        else:
            node_list = [tax_opt_node]
        strat = Strategy(node_list=node_list, ref_universe=ref_universe)
        sim_settings = SimulationSettings(calculation_type=calculation_type,
                                          from_date=from_date, to_date=to_date,
                                          analysis_date=analysis_date)

        super().__init__(
            strategy=strat,
            simulation_settings=sim_settings,
            profile_name=profile_name
        )


class InitialTETemplate(PureOptimizationTemplate):
    """
    Template to calculate tracking error of the initial portfolio

    Args:
        calculation_type (CalculationTypeEnum): (optional) Use to specify behavior of the calculation. The following options are available:
            
                    •REBALANCE - (default)rebalance a portfolio for one day only. Use the latest available portfolio given the date of calculation.
                    •SIMULATION - compute a time series of portfolio using the latest available portfolio given the first day of calculation. Skip last rebalancing.
                    •EOD/BACKCALCULATION - compute a time series of portfolio using the portfolio at the day of the calculation. Fail if not present. Perform last rebalancing.                    
        analysis_date (str): The date in YYYY-MM-DD format for which the account is being REBALANCED. Do not set with from_date.
        from_date (str): The starting date in YYYY-MM-DD format for a SIMULATION/ BACKCALCULATION. Do not set with analysis_date.
        to_date (str): The ending date in YYYY-MM-DD format for a SIMULATION/ BACKCALCULATION. Do not set with analysis_date.
        benchmark_list (list): List of MSCI index to use as a benchmark. Default value is None.
        universe_list (list): List of MSCI index to use as a universe. Default value is None.
        portfolio (Portfolio):  A portfolio object to use as the initial portfolio of a rebalance or simulation. Default value is CashPortfolio with 10,000,000 USD.
    """

    def __init__(self,
                 **kwargs):
        to_constraint = ConstraintFactory.TurnoverConstraint(0.0)
        super().__init__(
            risk_aversion=0,
            cash_lower_bound=0,
            cash_upper_bound=1e15,
            constraints=[to_constraint],
            **kwargs)


class UnrealizedGainsTemplate(GenericTaxTemplate):
    """
    Template to calculate unrealized gains of the initial portfolio

    Args:
        calculation_type (CalculationTypeEnum): (optional) Use to specify behavior of the calculation. The following options are available:
            
                    •REBALANCE - (default)rebalance a portfolio for one day only. Use the latest available portfolio given the date of calculation.
                    •SIMULATION - compute a time series of portfolio using the latest available portfolio given the first day of calculation. Skip last rebalancing.
                    •EOD/BACKCALCULATION - compute a time series of portfolio using the portfolio at the day of the calculation. Fail if not present. Perform last rebalancing.                    
        analysis_date (str): The date in YYYY-MM-DD format for which the account is being REBALANCED. Do not set with from_date.
        from_date (str): The starting date in YYYY-MM-DD format for a SIMULATION/ BACKCALCULATION. Do not set with analysis_date.
        to_date (str): The ending date in YYYY-MM-DD format for a SIMULATION/ BACKCALCULATION. Do not set with analysis_date.
        benchmark_list (list): List of MSCI index to use as a benchmark. Default value is None.
        universe_list (list): List of MSCI index to use as a universe. Default value is None.
        portfolio (Portfolio):  A portfolio object to use as the initial portfolio of a rebalance or simulation. Default value is CashPortfolio with 10,000,000 USD.
    """

    def __init__(self,
                 **kwargs):
        super().__init__(
            risk_aversion=0,
            tax_multiplier=1,
            loss_benefit_multiplier=0,
            cash_lower_bound=1,
            cash_upper_bound=1,
            **kwargs)
